//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-oci.rs: syd's OCI container runtime
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(unused_variables)]
#![allow(unused_assignments)]
#![allow(unused_mut)]

use std::{
    fs::{create_dir_all, remove_dir_all, File},
    path::{Path, PathBuf},
    process::ExitCode,
};

use anyhow::{bail, Context, Error, Result};
use getargs::{Opt, Options};
use nix::{
    errno::Errno,
    mount::{mount, MsFlags},
    sched::{unshare, CloneFlags},
    sys::stat::Mode,
    unistd::{chdir, mkdir, Pid, Uid},
};
use oci_spec::runtime::{LinuxNamespaceType, Spec};

fn main() -> Result<ExitCode> {
    syd::set_sigpipe_dfl().context("Failed to set SIGPIPE signal handler to default.")?;

    let uid = Uid::effective().as_raw();
    let mut root = if uid == 0 {
        String::from("/run/syd")
    } else if let Ok(path) = std::env::var("XDG_RUNTIME_DIR") {
        format!("{path}/syd")
    } else {
        format!("/run/syd-{uid}")
    };
    let mut opts = Options::new(
        argv::iter()
            .skip(1)
            .map(|os| os.to_str().expect("Argument is not valid UTF-8!")),
    );
    while let Some(opt) = opts
        .next_opt()
        .context("Failed to call main::Options::next!")?
    {
        match opt {
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('v') | Opt::Long("version") => {
                version();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Long("debug") => {}
            Opt::Long("systemd-cgroup") => {}
            Opt::Long("log") => {
                let _ = opts.value().context("--log requires an argument!")?;
            }
            Opt::Long("log-format") => {
                if !matches!(
                    opts.value().context("--log-format requires an argument!")?,
                    "text" | "json"
                ) {
                    bail!("Invalid log format!");
                }
            }
            Opt::Long("root") => {
                root = opts
                    .value()
                    .context("--root requires an argument!")?
                    .to_string();
            }
            Opt::Long("rootless") => {
                let _ = opts.value().context("--rootless requires an argument!")?;
            }
            _ => {
                bail!("Unknown option: {opt:?}!");
            }
        }
    }

    let root = Path::new(&root);
    match opts.next_positional() {
        Some("create") => cmd_create(opts, &root),
        Some("delete") => cmd_delete(opts, &root),
        Some("start") => cmd_start(opts),
        Some("state") => cmd_state(opts),
        Some("kill") => cmd_kill(opts),
        None => {
            help();
            Ok(ExitCode::SUCCESS)
        }
        Some(cmd) => bail!("Invalid subcommand {cmd}!"),
    }
}

fn cmd_create<I: Iterator<Item = &'static str>, P: AsRef<Path>>(
    mut opts: Options<&'static str, I>,
    root: P,
) -> Result<ExitCode> {
    let mut bundle = None;
    let mut socket = None;
    let mut pid_fn = None;
    while let Some(opt) = opts
        .next_opt()
        .context("Failed to call create::Options::next!")?
    {
        match opt {
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('b') | Opt::Long("bundle") => {
                bundle = Some(
                    opts.value()
                        .context("--bundle requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("console-socket") => {
                socket = Some(
                    opts.value()
                        .context("--console-socket requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("pid-file") => {
                pid_fn = Some(
                    opts.value()
                        .context("--pid-file requires an argument!")?
                        .to_string(),
                );
            }
            Opt::Long("no-new-keyring") => bail!("--no-new-keyring is not implemented yet!"),
            Opt::Long("no-pivot") => bail!("--no-pivot is insecure and therefore unsupported!"),
            Opt::Long("preserve-fds") => {
                bail!("--preserve-fds is insecure and therefore unsupported!")
            }
            _ => bail!("Unknown option: {opt:?}!"),
        }
    }

    let id = match opts.next_positional() {
        None => {
            help();
            return Ok(ExitCode::SUCCESS);
        }
        Some(id) => id,
    };

    if let Some(bundle) = bundle {
        chdir(Path::new(&bundle)).context("Failed to change to bundle directory!")?;
    }

    make_root(&root)?;
    make_home(&root, id)?;

    let spec = Spec::load("config.json").context("Failed to load config.json!")?;

    setup_namespaces(&spec, &root, &id).context("Failed to set up namespaces!")?;
    // TODO: next step after setup namespaces? cgroups? mounts? rlimits? etc.pp

    Ok(ExitCode::SUCCESS)
}

fn cmd_delete<I: Iterator<Item = &'static str>, P: AsRef<Path>>(
    mut opts: Options<&'static str, I>,
    root: P,
) -> Result<ExitCode> {
    let mut force = false;
    while let Some(opt) = opts
        .next_opt()
        .context("Failed to call delete::Options::next!")?
    {
        match opt {
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('f') | Opt::Long("force") => {
                force = true;
            }
            _ => bail!("Unknown option: {opt:?}!"),
        }
    }

    let id = match opts.next_positional() {
        None => {
            help();
            return Ok(ExitCode::SUCCESS);
        }
        Some(id) => id,
    };

    make_root(&root)?;
    remove_home(&root, id)?;

    Ok(ExitCode::SUCCESS)
}

fn cmd_start<I: Iterator<Item = &'static str>>(
    mut opts: Options<&'static str, I>,
) -> Result<ExitCode> {
    unimplemented!()
}

fn cmd_state<I: Iterator<Item = &'static str>>(
    mut opts: Options<&'static str, I>,
) -> Result<ExitCode> {
    unimplemented!()
}

fn cmd_kill<I: Iterator<Item = &'static str>>(
    mut opts: Options<&'static str, I>,
) -> Result<ExitCode> {
    unimplemented!()
}

fn help() {
    println!("Usage: syd-oci [OPTIONS] [COMMAND]");
    println!("Syd's OCI container runtime.");
    println!("Supported commands are create, start, state, kill and delete.");
}

fn version() {
    println!("syd-oci {}", *syd::config::VERSION);
    println!("Author: Ali Polatel");
    println!("License: GPL-3.0-or-later");
}

fn setup_namespaces<P: AsRef<Path>>(spec: &Spec, root: P, id: &str) -> Result<()> {
    if let Some(linux) = spec.linux() {
        if let Some(namespaces) = linux.namespaces() {
            let mut flags = CloneFlags::empty();
            let ns_dir_path = root.as_ref().join(id).join("ns");
            create_dir_all(&ns_dir_path)
                .context("Failed to create persistent namespace directory!")?;

            // Accumulate flags for all namespaces to be created.
            for ns in namespaces {
                match ns.typ() {
                    LinuxNamespaceType::Mount => flags.insert(CloneFlags::CLONE_NEWNS),
                    LinuxNamespaceType::Ipc => flags.insert(CloneFlags::CLONE_NEWIPC),
                    LinuxNamespaceType::Network => flags.insert(CloneFlags::CLONE_NEWNET),
                    LinuxNamespaceType::Pid => flags.insert(CloneFlags::CLONE_NEWPID),
                    LinuxNamespaceType::Uts => flags.insert(CloneFlags::CLONE_NEWUTS),
                    LinuxNamespaceType::User => flags.insert(CloneFlags::CLONE_NEWUSER),
                    LinuxNamespaceType::Cgroup => flags.insert(CloneFlags::CLONE_NEWCGROUP),
                    LinuxNamespaceType::Time => {} // TODO: Implement this when nix has CLONE_NEWTIME.
                }
            }

            // Only call unshare if there are flags set to avoid unnecessary system calls.
            if !flags.is_empty() {
                unshare(flags).context("Failed to create namespaces!")?;
            }
        }
    }

    Ok(())
}

fn remove_home<P: AsRef<Path>>(root: P, id: &str) -> Result<()> {
    if id.contains("/") {
        bail!("Invalid container ID!");
    }
    let path = root.as_ref().join(id);
    remove_dir_all(path).context("Failed to delete container home directory!")
}

fn make_home<P: AsRef<Path>>(root: P, id: &str) -> Result<()> {
    if id.contains("/") {
        bail!("Invalid container ID!");
    }
    let path = root.as_ref().join(id);
    mkdir(&path, Mode::S_IRWXU | Mode::S_ISVTX).context("Failed to create home directory!")
}

fn make_root<P: AsRef<Path>>(path: P) -> Result<()> {
    match mkdir(path.as_ref(), Mode::S_IRWXU | Mode::S_ISVTX) {
        Ok(_) | Err(Errno::EEXIST) => Ok(()),
        Err(errno) => Err(Error::new(errno).context("Failed to create root directory!")),
    }
}
